with Ada.Text_IO;

with Names;
with AocsMgr;
with FdirMgr;
with UnitMgr;
with ModeMgr.Transitions;


package body ModeMgr is

   use Names;


   procedure Init_Transition is
   begin

      Trans.Target := None;
      Trans.Done   := False;
      Trans.Step   := 0;
      Trans.Timer  := 0;

   end Init_Transition;


   Separation_Counter : Natural := 0;

   procedure Separation_StartCheck is
   begin

      if Separation_Counter = 0 then

         Separation_Counter := 1;

      end if;

   end Separation_StartCheck;


   function Separation_Done return Boolean is
      Ret_Value : Boolean;
   begin

      if Separation_Counter > 0 then

         if Separation_Counter > 10 then

            Ret_Value := True;
            return Ret_Value;

         end if;

         Separation_Counter := Separation_Counter + 1;

      end if;

      Ret_Value := False;
      return Ret_Value;

   end Separation_Done;


   procedure Start_Possible_Autonomous_Transition is
      Ret_Value_from_Call : Boolean;
   begin

      case Mode is

         when Off =>

            if not Trans.Standby_Entered_Once then
               -- Off entered after CDMU boot,
               -- force autonomous start of transition to Standby

               Trans.Standby_Entered_Once := True;

               Trans.Target := Standby;

               -- else
               -- Off entered from higher mode means
               -- that CDMU reboot is already requested
               -- and attitude control is stopped.
            end if;


         when Standby =>

            Ret_Value_from_Call := Separation_Done;
            if Ret_Value_from_Call then

               Trans.Target := Safe;

            end if;


         when Safe =>

            Ret_Value_from_Call := AocsMgr.Coarse_Pointing_Reached;
            if Ret_Value_from_Call then

               Trans.Target := Nominal;

            end if;


         when Nominal =>

            Ret_Value_from_Call := AocsMgr.Fine_Pointing_Reached;
            if Ret_Value_from_Call then

               Trans.Target := Preparation;

            end if;


         when Preparation =>

            Ret_Value_from_Call := AocsMgr.PLI_Is_Ready;
            if Ret_Value_from_Call then

               Trans.Target := Science;

            end if;


         when Science =>

            null;
            -- Stay in Science mode until FDIR forces back to some
            -- earlier mode

      end case;

   end Start_Possible_Autonomous_Transition;


   procedure Start_FDIR_Transition (Target_Mode : Target_Mode_T) is
   begin

      -- Abort possible on-going transition:
      Init_Transition;

      Trans.FDIR_Request := Target_Mode;

      Trans.Target := Target_Mode;

   end Start_FDIR_Transition;


   procedure Set_FDIR_Request (Target_Mode : Target_Mode_T)
   is
      -- Invariant: Target_Mode /= Standby and Target_Mode /= None
      -- Invariant: Target_Mode < Mode
      --            or (Mode = Safe and Target_Mode = Safe)
      Chosen_Target_Mode : Target_Mode_T;
   begin

      if Trans.FDIR_Request = None then
         -- No on-going old FDIR request

         Chosen_Target_Mode := Target_Mode;

      -- Error while recovery is already running:
      elsif Target_Mode = Off
         or Trans.Target = Off
         or Trans.Target = Standby
         or Trans.Target = Safe
         or Trans.FDIR_Request = Off
         or Trans.FDIR_Request = Safe
      then

         Chosen_Target_Mode := Off;

      else

         Chosen_Target_Mode := Safe;

      end if;

      Start_FDIR_Transition (Chosen_Target_Mode);

   end Set_FDIR_Request;


   procedure Request_CDMU_Reboot is
   begin

      AocsMgr.Select_Controller (AocsMgr.None);

      AocsMgr.Stop_PLI_Management;

      UnitMgr.Release_All;

      Trans.Emergency_Reboot_Requested := True;

   end Request_CDMU_Reboot;


   procedure Execute_Step is
   begin

      case Trans.Target is

         when Off =>
            -- Emergency transition to Off

            Request_CDMU_Reboot;
            Trans.Done := True;


         when Standby =>

            Transitions.To_StandbyMode;

         when Safe =>
            Transitions.To_SafeMode;


         when Nominal =>

            if Mode = Safe then

               Transitions.To_NominalMode;

            elsif Mode /= Off
               and Mode /= Standby
            then

               Transitions.Back_To_Nominal;

            end if;


         when Preparation =>

            Transitions.To_PrepMode;


         when Science =>

            Transitions.To_ScienceMode;

         when None =>

            null;

      end case;

   end Execute_Step;


   procedure Main
   is
      Prev_Step : Natural;
   begin

      if Trans.Target = None then
         -- No old transition on-going, and no new requests
         --
         -- Evaluate whether an autonomous transition should be started

         Start_Possible_Autonomous_Transition;

      end if;

      Ada.Text_IO.Put_Line ("trtarg="&Trans.Target'Img);
      if Trans.Target /= None then

         Prev_Step := Trans.Step;

         Execute_Step;
         -- Selects the procedure corresponding Trans.Current.
         -- The transition step counter is separately maintained.
         -- Here we continue execution of a transition.
         -- The step condition is always evaluated, and if it is true,
         -- the step activities are performed and step counter updated.

         if Trans.Done then
            -- If the last step is reached, time to change the mode

            Mode := Trans.Target;
            -- The resulting mode can be deduced from the transition (name)

            Init_Transition;

            Trans.FDIR_Request := None;

         elsif Prev_Step < Trans.Step then
            -- We have progress

            Trans.Timer := 0;

         else

            Trans.Timer := Trans.Timer + 1;

            if Trans.Timer > Trans.Timeout_Limit then

               Init_Transition;

               Trans.Error := Timeout;

            end if;

         end if;

      end if;

   end Main;


end ModeMgr;
